@@ -16,15 +16,20 @@
 
    Note, this is not iSCSI.
 """
+
+from __future__ import annotations
+
 import glob
 import os
 import re
 import time
-from typing import Dict, List, Optional  # noqa: H301
+from typing import Optional
 
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
+from oslo_utils import excutils
 
+from os_brick import constants
 from os_brick import exception
 from os_brick import executor
 from os_brick.privileged import rootwrap as priv_rootwrap
@@ -37,12 +42,77 @@ MULTIPATH_WWID_REGEX = re.compile(r"\((?P<wwid>.+)\)")
 MULTIPATH_DEVICE_ACTIONS = ['unchanged:', 'reject:', 'reload:',
                             'switchpg:', 'rename:', 'create:',
                             'resize:']
+MULTIPATHD_RESIZE_TIMEOUT = 120
 
 
 class LinuxSCSI(executor.Executor):
     # As found in drivers/scsi/scsi_lib.c
     WWN_TYPES = {'t10.': '1', 'eui.': '2', 'naa.': '3'}
 
+    @staticmethod
+    def lun_for_addressing(lun, addressing_mode=None):
+        """Convert luns to values used by the system.
+
+        How a LUN is codified depends on the standard being used by the storage
+        array and the mode, which is unknown by the host.
+
+        Addressing modes based on the standard:
+            * SAM:
+              - 64bit address
+
+            * SAM-2:
+              - Peripheral device addressing method (Code 00b)
+                + Single level
+                + Multi level
+              - Flat space addressing method (Code 01b)
+              - Logical unit addressing mode (Code 10b)
+              - Extended logical unit addressing method (Code 11b)
+
+            * SAM-3: Mostly same as SAM-2 but with some differences,
+              like supporting addressing LUNs < 256 with flat address space.
+
+        This means that the same LUN numbers could have different addressing
+        values.  Examples:
+          * LUN 1:
+            - SAM representation: 1
+            - SAM-2 peripheral: 1
+            - SAM-2 flat addressing: Invalid
+            - SAM-3 flat addressing: 16384
+
+          * LUN 256
+            - SAM representation: 256
+            - SAM-2 peripheral: Not possible to represent
+            - SAM-2 flat addressing: 16640
+            - SAM-3 flat addressing: 16640
+
+        This method makes the transformation from the numerical LUN value to
+        the right addressing value based on the addressing_mode.
+
+        Acceptable values are:
+        - SAM: 64bit address with no translation
+        - transparent: Same as SAM but used by drivers that want to use non
+                       supported addressing modes by using the addressing mode
+                       instead of the LUN without being misleading (untested).
+        - SAM2: Peripheral for LUN < 256 and flat for LUN >= 256. In SAM-2
+                flat cannot be used for 0-255
+        - SAM3-flat: Force flat-space addressing
+
+        The default is SAM/transparent and nothing will be done with the LUNs.
+        """
+        mode = addressing_mode or constants.SCSI_ADDRESSING_SAM
+        if mode not in constants.SCSI_ADDRESSING_MODES:
+            raise exception.InvalidParameterValue('Invalid addressing_mode '
+                                                  f'{addressing_mode}')
+
+        if (mode == constants.SCSI_ADDRESSING_SAM3_FLAT
+                or (mode == constants.SCSI_ADDRESSING_SAM2 and lun >= 256)):
+            old_lun = lun
+            lun += 16384
+            LOG.info('Transforming LUN value for addressing: %s -> %s',
+                     old_lun, lun)
+
+        return lun
+
     def echo_scsi_command(self, path, content) -> None:
         """Used to echo strings to scsi subsystem."""
 
@@ -80,7 +150,7 @@ class LinuxSCSI(executor.Executor):
             with exc.context(force, 'Removing %s failed', device):
                 self.echo_scsi_command(path, "1")
 
-    def wait_for_volumes_removal(self, volumes_names: List[str]) -> None:
+    def wait_for_volumes_removal(self, volumes_names: list[str]) -> None:
         """Wait for device paths to be removed from the system."""
         str_names = ', '.join(volumes_names)
         LOG.debug('Checking to see if SCSI volumes %s have been removed.',
@@ -103,7 +173,7 @@ class LinuxSCSI(executor.Executor):
                     LOG.debug('%s still exist.', ', '.join(exist))
         raise exception.VolumePathNotRemoved(volume_path=exist)
 
-    def get_device_info(self, device: str) -> Dict[str, Optional[str]]:
+    def get_device_info(self, device: str) -> dict[str, Optional[str]]:
         dev_info = {'device': device, 'host': None,
                     'channel': None, 'id': None, 'lun': None}
         # The input argument 'device' can be of 2 types:
@@ -130,7 +200,7 @@ class LinuxSCSI(executor.Executor):
         LOG.debug('dev_info=%s', str(dev_info))
         return dev_info
 
-    def get_sysfs_wwn(self, device_names, mpath=None) -> str:
+    def get_sysfs_wwn(self, device_names: list[str], mpath=None) -> str:
         """Return the wwid from sysfs in any of devices in udev format."""
         # If we have a multipath DM we know that it has found the WWN
         if mpath:
@@ -152,7 +222,7 @@ class LinuxSCSI(executor.Executor):
             return wwid
 
         # If we have multiple designators use symlinks to find out the wwn
-        device_names = set(device_names)
+        device_names_set = set(device_names)
         for wwn_path in wwn_paths:
             try:
                 if os.path.islink(wwn_path) and os.stat(wwn_path):
@@ -168,11 +238,11 @@ class LinuxSCSI(executor.Executor):
                             dm_devs = os.listdir(slaves_path)
                             # This is the right wwn_path if the devices we have
                             # attached belong to the dm we followed
-                            if device_names.intersection(dm_devs):
+                            if device_names_set.intersection(dm_devs):
                                 break
 
                         # This is the right wwn_path if  devices we have
-                        elif name in device_names:
+                        elif name in device_names_set:
                             break
             except OSError:
                 continue
@@ -195,7 +265,7 @@ class LinuxSCSI(executor.Executor):
             return udev_wwid
         return ''
 
-    def get_scsi_wwn(self, path):
+    def get_scsi_wwn(self, path: str) -> str:
         """Read the WWN from page 0x83 value for a SCSI device."""
 
         (out, _err) = self._execute('/lib/udev/scsi_id', '--page', '0x83',
@@ -547,17 +617,6 @@ class LinuxSCSI(executor.Executor):
             return info
         return None
 
-    def get_device_size(self, device):
-        """Get the size in bytes of a volume."""
-        (out, _err) = self._execute('blockdev', '--getsize64',
-                                    device, run_as_root=True,
-                                    root_helper=self._root_helper)
-        var = str(out.strip())
-        if var.isnumeric():
-            return int(var)
-        else:
-            return None
-
     def multipath_reconfigure(self):
         """Issue a multipathd reconfigure.
 
@@ -571,23 +630,64 @@ class LinuxSCSI(executor.Executor):
                                     root_helper=self._root_helper)
         return out
 
-    def multipath_resize_map(self, mpath_id):
+    def _multipath_resize_map(self, dm_path):
+        cmd = ('multipathd', 'resize', 'map', dm_path)
+        (out, _err) = self._execute(*cmd,
+                                    run_as_root=True,
+                                    root_helper=self._root_helper)
+        if 'fail' in out or 'timeout' in out:
+            raise putils.ProcessExecutionError(
+                stdout=out, stderr=_err,
+                exit_code=1, cmd=cmd)
+
+        return out
+
+    def multipath_resize_map(self, dm_path):
         """Issue a multipath resize map on device.
 
         This forces the multipath daemon to update it's
         size information a particular multipath device.
-        """
-        (out, _err) = self._execute('multipathd', 'resize', 'map', mpath_id,
-                                    run_as_root=True,
-                                    root_helper=self._root_helper)
-        return out
 
-    def extend_volume(self, volume_paths, use_multipath=False):
+        :param dm_path: Real path of the DM device (eg: /dev/dm-5)
+        """
+
+        # "multipathd reconfigure" is async since 0.6.1. While the
+        # operation is in progress, "multipathd resize map" returns
+        # "timeout".
+        tstart = time.time()
+        while True:
+            try:
+                self._multipath_resize_map(dm_path)
+                break
+            except putils.ProcessExecutionError as err:
+                with excutils.save_and_reraise_exception(reraise=True) as ctx:
+                    elapsed = time.time() - tstart
+                    if 'timeout' in err.stdout and (
+                            elapsed < MULTIPATHD_RESIZE_TIMEOUT):
+                        LOG.debug(
+                            "multipathd resize map timed out. "
+                            "Elapsed: %s, timeout: %s. Retrying...",
+                            elapsed, MULTIPATHD_RESIZE_TIMEOUT)
+                        ctx.reraise = False
+                        time.sleep(1)
+
+    def extend_volume(self,
+                      volume_paths: list,
+                      use_multipath: bool = False) -> Optional[int]:
         """Signal the SCSI subsystem to test for volume resize.
 
         This function tries to signal the local system's kernel
         that an already attached volume might have been resized.
         """
+        # We need all paths up before extending the devices.
+        # see Launchpad Bug: #2032177 for more details.
+        LOG.debug("Checking paths are valid %s", volume_paths)
+        for volume_path in volume_paths:
+            if not utils.check_valid_device(self, volume_path):
+                LOG.error("Path status is down for path %s", volume_path)
+                raise exception.BrickException("All paths need to be up "
+                                               "to extend the device.")
+
         LOG.debug("extend volume %s", volume_paths)
 
         for volume_path in volume_paths:
@@ -602,13 +702,13 @@ class LinuxSCSI(executor.Executor):
             scsi_path = ("/sys/bus/scsi/drivers/sd/%(device_id)s" %
                          {'device_id': device_id})
 
-            size = self.get_device_size(volume_path)
+            size = utils.get_device_size(self, volume_path)
             LOG.debug("Starting size: %s", size)
 
             # now issue the device rescan
             rescan_path = "%(scsi_path)s/rescan" % {'scsi_path': scsi_path}
             self.echo_scsi_command(rescan_path, "1")
-            new_size = self.get_device_size(volume_path)
+            new_size = utils.get_device_size(self, volume_path)
             LOG.debug("volume size after scsi device rescan %s", new_size)
 
         scsi_wwn = self.get_scsi_wwn(volume_paths[0])
@@ -618,18 +718,13 @@ class LinuxSCSI(executor.Executor):
                 # Force a reconfigure so that resize works
                 self.multipath_reconfigure()
 
-                size = self.get_device_size(mpath_device)
+                size = utils.get_device_size(self, mpath_device)
                 LOG.info("mpath(%(device)s) current size %(size)s",
                          {'device': mpath_device, 'size': size})
-                result = self.multipath_resize_map(scsi_wwn)
-                if 'fail' in result:
-                    LOG.error("Multipathd failed to update the size mapping "
-                              "of multipath device %(scsi_wwn)s volume "
-                              "%(volume)s",
-                              {'scsi_wwn': scsi_wwn, 'volume': volume_paths})
-                    return None
 
-                new_size = self.get_device_size(mpath_device)
+                self.multipath_resize_map(os.path.realpath(mpath_device))
+
+                new_size = utils.get_device_size(self, mpath_device)
                 LOG.info("mpath(%(device)s) new size %(size)s",
                          {'device': mpath_device, 'size': new_size})
 
@@ -740,3 +835,21 @@ class LinuxSCSI(executor.Executor):
                                        check_exit_code=False,
                                        root_helper=self._root_helper)
         return stdout.strip() == 'ok'
+
+    @utils.retry((putils.ProcessExecutionError, exception.BrickException),
+                 retries=3)
+    def multipath_del_map(self, mpath):
+        """Stop monitoring a multipath given its device name (eg: dm-7).
+
+        Method ensures that the multipath device mapper actually dissapears
+        from sysfs.
+        """
+        map_name = self.get_dm_name(mpath)
+        if map_name:
+            self._execute('multipathd', 'del', 'map', map_name,
+                          run_as_root=True, timeout=5,
+                          root_helper=self._root_helper)
+
+        if map_name and self.get_dm_name(mpath):
+            raise exception.BrickException("Multipath doesn't go away")
+        LOG.debug('Multipath %s no longer present', mpath)
